local MAJOR = "LibUnitID-1.0"
--@alpha@
MAJOR = MAJOR.."-alpha"
--@end-alpha@
local MINOR = 14
local LibUnitID,oldminor = LibStub:NewLibrary(MAJOR,MINOR)
if (not LibUnitID) then return end -- already loaded, no upgrade necessary

-- localize upvalues
local error,next,pairs,setmetatable,type,xpcall = error,next,pairs,setmetatable,type,xpcall
local geterrorhandler,wipe = geterrorhandler,wipe
local API_UnitGUID = UnitGUID

-- dump data from old version, if any
local olddata = {}
if (type(LibUnitID.__dump_for_upgrade) == "function") then
	if (oldminor < 9) then
		local frame,listeners,tests,updates = LibUnitID:__dump_for_upgrade()
		olddata = { frame=frame, tagListeners=listeners, logTests=tests, logUpdates=updates }
	else
		olddata = LibUnitID:__dump_for_upgrade()
	end
	olddata.MINOR = oldminor
end


--[[ STATIC HASH TABLES ]]--

-- tags that can be tracked with events
local tagSupported = {
	--player=true,
	vehicle=true,vehicletarget=true,
	pet=true,pettarget=true,
	target=true,targettarget=true,
	focus=true,focustarget=true
}
for i=1,5 do
	tagSupported["party"..i] = true
	tagSupported["party"..i.."target"] = true
	tagSupported["partypet"..i] = true
	tagSupported["partypet"..i.."target"] = true
end
for i=1,40 do
	tagSupported["raid"..i] = true
	tagSupported["raid"..i.."target"] = true
	tagSupported["raidpet"..i] = true
	tagSupported["raidpet"..i.."target"] = true
end

-- tags that are the pet of (thus dependent on) other tags
local tagPet = {
	player="pet"
	--player="vehicle" is equally valid but doesn't fit the scheme very well, so there's a special check for that during UNIT_PET
	--target and focus get UNIT_PET events, but targetpet and focuspet aren't valid unitIDs.  boggle
}
for i=1,5 do
	tagPet["party"..i] = "partypet"..i
end
for i=1,40 do
	tagPet["raid"..i] = "raidpet"..i
end

-- tags that are the target of (thus dependent on) other tags
local tagTarget = {
	--player="target" not needed since we use PLAYER_TARGET_CHANGED instead of UNIT_TARGET(player)
	vehicle="vehicletarget",
	pet="pettarget",
	target="targettarget",
	focus="focustarget"
}
for i=1,5 do
	tagTarget["party"..i] = "party"..i.."target"
	tagTarget["partypet"..i] = "partypet"..i.."target"
end
for i=1,40 do
	tagTarget["raid"..i] = "raid"..i.."target"
	tagTarget["raidpet"..i] = "raidpet"..i.."target"
end

-- tags that depend on another tag
local tagParent = {
	--target="player" not needed since player never changes
	--vehicle="player", ditto
	vehicletarget="vehicle",
	pettarget="pet",
	targettarget="target",
	focustarget="focus"
}
for i=1,5 do
	tagParent["party"..i.."target"] = "party"..i
	tagParent["partypet"..i] = "party"..i
	tagParent["partypet"..i.."target"] = "partypet"..i
end
for i=1,40 do
	tagParent["raid"..i.."target"] = "raid"..i
	tagParent["raidpet"..i] = "raid"..i
	tagParent["raidpet"..i.."target"] = "raidpet"..i
end

-- tags that depend on the party composition
local partyTags = {}
for i=1,5 do
	partyTags["party"..i] = true
	partyTags["party"..i.."target"]=true
	partyTags["partypet"..i] = true
	partyTags["partypet"..i.."target"] = true
end

-- tags that depend on the raid composition
local raidTags = {}
for i=1,40 do
	raidTags["raid"..i] = true
	raidTags["raid"..i.."target"] = true
	raidTags["raidpet"..i] = true
	raidTags["raidpet"..i.."target"] = true
end


--[[ SHARED RUNTIME DATA ]]--

local tablePool = {}
local hashPool = {}
local arrayPool = {}
local frame = olddata.frame or CreateFrame("Frame", nil, UIParent)
local tagTracked = {} -- tagTracked[unitid] = true/false
local tagID = {} -- tagID[unitid] = guid
local tagFunctions = {} -- tagFunctions[unitid][cb] = true
local tagStaticMethods = {} -- tagStaticMethods[unitid][handler][cb:func] = true
local tagDynamicMethods = {} -- tagDynamicMethods[unitid][handler][cb:str] = true
local queueFunctions = {}
local queueMethods = {}
local callbackEvent
local callbackFunction
local callbackMethod
local callbackHandler
local eventHead = 0
local eventTail = 0
local eventQueue = {}
local eventArg1 = {}
local logActive = false
local logTests = olddata.logTests or 0
local logUpdates = olddata.logUpdates or 0
local logQueries = olddata.logQueries or 0


--[[ DELTA TABLE ]]--

--- The "delta" table is provided to all callbacks on any UnitID change.
-- @name delta
-- @class table
-- This table lists the UnitIDs that changed during a single game event.  For
-- efficiency, the table is shared and provided to every callback involved
-- during a game event; for safety, that table may not be modified (it
-- '''__index'''es the real table, and errors on '''__newindex''').  This
-- means you cannot process it using '''pairs()''', '''ipairs()''' or
-- '''next()'''.  Instead, index '''0''' of the table will contain the number
-- of affected UnitIDs (the equivalent of the '''#''' operator), and indecies
-- '''1'''-'''#''' will contain those UnitIDs.  In addition, each affected
-- UnitID is a string key of the table, whose value is the boolean '''true'''.
-- Please see the project documentation for code examples to process this
-- table in your callback functions.
-- @field 0 The number of UnitIDs that changed.
-- @field (n) (Integer) For n = 1 .. delta[0], delta[n] is a UnitID that changed.
-- @field (unitid) (String) For unitid in { delta[1] .. delta[delta[0]] }, delta[unitid]=true.
local delta = {}
local deltaTail = 0
local deltaMeta = {
	__index = delta,
	__newindex = function() error(MAJOR..": cannot modify shared table") end,
	__metatable = false
}
local deltaWrapper = setmetatable({}, deltaMeta)


--[[ CALLBACK (UN)REGISTRATION ]]--

local function update_events()
	local pairs = pairs
	local tagTracked,tagPet,tagTarget,partyTags,raidTags = tagTracked,tagPet,tagTarget,partyTags,raidTags
	-- identify necessary events
	local p_t_c = (tagTracked.target) -- PLAYER_TARGET_CHANGED
	local p_f_c = (tagTracked.focus) -- PLAYER_FOCUS_CHANGED
	local u_p = (tagTracked.vehicle or tagTracked.vehicletarget or tagTracked.pet) -- UNIT_PET
	if (not u_p) then
		for parent,tag in pairs(tagPet) do
			if (tagTracked[tag]) then
				u_p = true
				break
			end
		end
	end
	local u_t = (tagTracked.targettarget or tagTracked.focustarget or tagTracked.pettarget) -- UNIT_TARGET
	if (not u_t) then
		for parent,tag in pairs(tagTarget) do
			if (tagTracked[tag]) then
				u_t = true
				break
			end
		end
	end
	local p_m_c = false -- PARTY_MEMBERS_CHANGED
	for tag,_ in pairs(partyTags) do
		if (tagTracked[tag]) then
			p_m_c = true
			break
		end
	end
	local r_r_u = false -- RAID_ROSTER_UPDATE
	for tag,_ in pairs(raidTags) do
		if (tagTracked[tag]) then
			r_r_u = true
			break
		end
	end
	-- register necessary events
	local frame = frame
	frame:UnregisterAllEvents()
	if (p_t_c) then frame:RegisterEvent("PLAYER_TARGET_CHANGED") end
	if (p_f_c) then frame:RegisterEvent("PLAYER_FOCUS_CHANGED") end
	if (u_p) then frame:RegisterEvent("UNIT_PET") end
	if (u_t) then frame:RegisterEvent("UNIT_TARGET") end
	if (p_m_c) then frame:RegisterEvent("PARTY_MEMBERS_CHANGED") end
	if (r_r_u) then frame:RegisterEvent("RAID_ROSTER_UPDATE") end
end -- update_events()

local function add_tag_listener(tag, cb, handler)
	if (not tagSupported[tag]) then
		return false
	end
	-- validate args and store the callback
	local tF,tSM,tDM = tagFunctions,tagStaticMethods,tagDynamicMethods
	local first = false
	if (type(handler) == "table") then
		-- method
		if (type(cb) == "function") then
			-- static method
			if (tSM[tag] and tSM[tag][handler] and tSM[tag][handler][cb]) then
				return true
			end
			tSM[tag] = tSM[tag] or next(hashPool) or {}
			hashPool[tSM[tag]] = nil
			if (not tSM[tag][handler]) then
				first = not (tF[tag] or tDM[tag])
				tSM[tag][handler] = next(hashPool) or {}
				hashPool[tSM[tag][handler]] = nil
			end
			tSM[tag][handler][cb] = true
		elseif (type(cb) == "string") then
			-- dynamic method
			if (tDM[tag] and tDM[tag][handler] and tDM[tag][handler][cb]) then
				return true
			end
			tDM[tag] = tDM[tag] or next(hashPool) or {}
			hashPool[tDM[tag]] = nil
			if (not tDM[tag][handler]) then
				first = not (tF[tag] or tSM[tag])
				tDM[tag][handler] = next(hashPool) or {}
				hashPool[tDM[tag][handler]] = nil
			end
			tDM[tag][handler][cb] = true
		else
			return false -- invalid callback type
		end
		queueMethods[handler] = queueMethods[handler] or next(hashPool) or {}
		hashPool[queueMethods[handler]] = nil
	elseif (type(cb) == "function") then
		-- function
		if (tF[tag] and tF[tag][cb]) then
			return true
		end
		if (not tF[tag]) then
			first = not (tSM[tag] or tDM[tag])
			tF[tag] = next(hashPool) or {}
			hashPool[tF[tag]] = nil
		end
		tF[tag][cb] = true
	else
		return false -- invalid callback type
	end
	-- if this is the tag's first listener, establish identities
	if (first) then
		tagTracked[tag] = true
		tagID[tag] = API_UnitGUID(tag)
		local parent = tagParent[tag]
		while (parent) do
			tagTracked[parent] = true
			tagID[parent] = tagID[parent] or API_UnitGUID(parent)
			parent = tagParent[parent]
		end
		update_events()
	end
	return true
end -- add_tag_listener()

local update_tag_watch -- symbol must be defined before the function so that it can recurse
update_tag_watch = function(tag)
	-- recurse to child tags first, since one of them might get cleaned up even if we still have a listener
	-- don't need player=>vehicle because we'll never call this on player anyway
	local pet_watched = (tagPet[tag] and update_tag_watch(tagPet[tag]))
	local target_watched = (tagTarget[tag] and update_tag_watch(tagTarget[tag]))
	-- but if nobody's listening, stop watching
	if (not (tagFunctions[tag] or tagStaticMethods[tag] or tagDynamicMethods[tag] or pet_watched or target_watched)) then
		tagTracked[tag] = false
		tagID[tag] = nil
		return false
	end
	return true
end -- update_tag_watch()

local function remove_tag_listener(tag, cb, handler)
	-- validate args and erase the callback
	local tF,tSM,tDM = tagFunctions,tagStaticMethods,tagDynamicMethods
	local last = false
	if (type(handler) == "table") then
		-- method
		if (type(cb) == "function") then
			-- static method
			if (tSM[tag] and tSM[tag][handler]) then
				tSM[tag][handler][cb] = nil
				-- if this was the last static method for this handler for this tag, clean up
				if (not next(tSM[tag][handler])) then
					hashPool[tSM[tag][handler]] = true
					tSM[tag][handler] = nil
					-- if this was the last handler for this tag, clean up
					if (not next(tSM[tag])) then
						last = not (tF[tag] or tDM[tag])
						hashPool[tSM[tag]] = true
						tSM[tag] = nil
					end
				end
			end
		elseif (type(cb) == "string") then
			-- dynamic method
			if (tDM[tag] and tDM[tag][handler]) then
				tDM[tag][handler][cb] = nil
				-- if this was the last dynamic method for this handler for this tag, clean up
				if (not next(tDM[tag][handler])) then
					hashPool[tDM[tag][handler]] = true
					tDM[tag][handler] = nil
					-- if this was the last handler for this tag, clean up
					if (not next(tDM[tag])) then
						last = not (tF[tag] or tSM[tag])
						hashPool[tDM[tag]] = true
						tDM[tag] = nil
					end
				end
			end
		else
			return false -- invalid callback type
		end
		-- if handler is no longer used anywhere, recycle it in queueMethods
		local used = false
		for _,handlers in pairs(tSM) do
			for _h,_ in pairs(handlers) do
				if (_h == handler) then
					used = true
					break
				end
			end
			if (used) then
				break
			end
		end
		if (not used) then
			for _,handlers in pairs(tDM) do
				for _h,_ in pairs(handlers) do
					if (_h == handler) then
						used = true
						break
					end
				end
				if (used) then
					break
				end
			end
			if (not used) then
				hashPool[queueMethods[handler]] = true
				queueMethods[handler] = nil
			end
		end
	elseif (type(cb) == "function") then
		-- function
		if (tF[tag]) then
			tF[tag][cb] = nil
			-- if this was the last function for this tag, clean up
			if (not next(tF[tag])) then
				last = not (tSM[tag] or tDM[tag])
				hashPool[tF[tag]] = true
				tF[tag] = nil
			end
		end
	else
		return false -- invalid callback type
	end
	-- if this was the tag's last listener, clean up
	if (last) then
		while (tagParent[tag]) do
			tag = tagParent[tag]
		end
		update_tag_watch(tag)
		update_events()
	end
	return true
end -- remove_tag_listener()


--[[ CACHE UPDATE ]]--

local update_unit -- symbol must be defined before the function so that it can recurse
update_unit = function(tag, delta)
	-- NOTE: any change to this function must be mirrored in compile_update_unit_tests(), for accurate profiling
	if (not tagTracked[tag]) then return end
	local newID = API_UnitGUID(tag)
	if (newID == tagID[tag]) then return end
	-- record the change
	tagID[tag] = newID
	deltaTail = deltaTail + 1
	delta[deltaTail] = tag
	-- cascade to child tags, maximizing tail calls
	-- don't need player=>vehicle because we'll never call this on player anyway
	if (tagTarget[tag]) then
		if (tagPet[tag]) then
			update_unit(tagPet[tag], delta)
		end
		return update_unit(tagTarget[tag], delta)
	elseif (tagPet[tag]) then
		return update_unit(tagPet[tag], delta)
	end
end -- update_unit()

local function compile_update_unit_tests()
	-- return two test functions: one where the update short circuits on no-change, and one where it doesn't
	local testFunc
	do
		local tagTracked = { target=1 }
		local API_UnitGUID = API_UnitGUID
		local tagID = {}
		testFunc = function(tag) -- thou shalt only call this with the argument "target"!
			if (not tagTracked[tag]) then return true end
			local newID = API_UnitGUID(tag)
			if (newID == tagID[tag]) then return newID end
			tagID[tag] = newID -- this is only done once, so in a loop of thousands should have negligable affect on the total time
		end
	end
	local updateFunc
	do
		local tagTracked = { target=1 }
		local API_UnitGUID = API_UnitGUID
		local tagID = { target=1 }
		local _tagID = { target=1 }
		local delta = { [1]="target" }
		local deltaTail = 1
		local _deltaTail = 1
		local tagTarget = {}
		local tagPet = {}
		updateFunc = function(tag)
			if (not tagTracked[tag]) then return end
			local newID = API_UnitGUID(tag)
			if (newID == tagID[tag]) then return end
			_tagID[tag] = newID
			_deltaTail = _deltaTail + 1
			delta[deltaTail] = tag
			-- recursive calls are still logged, so we only need to measure if...elseif...end overhead here
			if (tagTarget[tag]) then error() elseif (tagPet[tag]) then error() end
		end
	end
	return testFunc,updateFunc
end -- compile_update_unit_tests()


--[[ CALLBACK DISPATCH ]]--

local function dispatch_function_callback()
	return callbackFunction(callbackEvent, deltaWrapper)
end -- dispatch_function_callback()

local function dispatch_method_callback()
	return callbackMethod(callbackHandler, callbackEvent, deltaWrapper)
end -- dispatch_method_callback()

local function invalid_method_callback()
	error(MAJOR..": failed to resolve callback method: "..callbackMethod)
end -- invalid_method_callback()

local function handle_callback_error(err)
	return geterrorhandler()(err)
end -- handle_callback_error()

local process_event -- function defined below, but needed inside dispatch_callbacks()

local function dispatch_callbacks(event)
	local next,pairs,type,wipe,xpcall = next,pairs,type,wipe,xpcall
	local dispatch_function_callback = dispatch_function_callback
	local dispatch_method_callback = dispatch_method_callback
	local handle_callback_error = handle_callback_error
	local tagFunctions,tagStaticMethods,tagDynamicMethods = tagFunctions,tagStaticMethods,tagDynamicMethods
	local queueFunctions,queueMethods = queueFunctions,queueMethods
	-- add size and string keys to delta table, and find applicable callbacks
	local tag,tF,tSM,tDM,qMh
	local delta,dT = delta,deltaTail
	delta[0] = dT
	for i = 1,dT do
		tag = delta[i]
		delta[tag] = true
		tF = tagFunctions[tag]
		if (tF) then
			for cb,_ in pairs(tF) do
				queueFunctions[cb] = true
			end
		end
		tSM = tagStaticMethods[tag]
		if (tSM) then
			for handler,cbs in pairs(tSM) do
				qMh = queueMethods[handler]
				for cb,_ in pairs(cbs) do
					qMh[cb] = true
				end
			end
		end
		tDM = tagDynamicMethods[tag]
		if (tDM) then
			for handler,methods in pairs(tDM) do
				qMh = queueMethods[handler]
				for method,_ in pairs(methods) do
					if (type(handler[method]) == "function") then
						qMh[handler[method]] = true
					else
						callbackMethod = method
						xpcall(invalid_method_callback, handle_callback_error)
					end
				end
			end
		end
	end
	callbackEvent = event
	-- run function callbacks
	for cb,_ in pairs(queueFunctions) do
		callbackFunction = cb
		xpcall(dispatch_function_callback, handle_callback_error)
	end
	wipe(queueFunctions)
	-- run method callbacks
	for handler,cbs in pairs(queueMethods) do
		for cb,_ in pairs(cbs) do
			callbackMethod = cb
			callbackHandler = handler
			xpcall(dispatch_method_callback, handle_callback_error)
		end
		wipe(cbs)
	end
	-- clean up
	callbackEvent = nil
	callbackFunction = nil
	callbackMethod = nil
	callbackHandler = nil
	wipe(delta)
	deltaTail = 0
	-- if the event queue is now empty, reset it; otherwise, process the next one
	if (eventHead == eventTail) then
		eventHead = 0
		eventTail = 0
		wipe(eventQueue)
		wipe(eventArg1)
	else
		eventHead = eventHead + 1
		return process_event(eventQueue[eventHead], eventArg1[eventHead])
	end
end -- dispatch_callbacks()


--[[ EVENT HANDLER ]]--

process_event = function(event, arg1)
	-- update affected unit(s)
	local delta = delta
	if (event == "UNIT_TARGET") then
		update_unit(tagTarget[arg1] or false, delta)
	elseif (event == "PLAYER_TARGET_CHANGED") then
		update_unit("target", delta)
	elseif (event == "PLAYER_FOCUS_CHANGED") then
		update_unit("focus", delta)
	elseif (event == "PARTY_MEMBERS_CHANGED") then
		local uu,d = update_unit,delta
		uu("party1",d); uu("party2",d); uu("party3",d); uu("party4",d); uu("party5",d)
	elseif (event == "RAID_ROSTER_UPDATE") then
		local uu,d = update_unit,delta
		uu("raid1", d); uu("raid2", d); uu("raid3", d); uu("raid4", d); uu("raid5", d)
		uu("raid6", d);  u("raid7", d); uu("raid8", d); uu("raid9", d); uu("raid10",d)
		uu("raid11",d); uu("raid12",d); uu("raid13",d); uu("raid14",d); uu("raid15",d)
		uu("raid16",d); uu("raid17",d); uu("raid18",d); uu("raid19",d); uu("raid20",d)
		uu("raid21",d); uu("raid22",d); uu("raid23",d); uu("raid24",d); uu("raid25",d)
		uu("raid26",d); uu("raid27",d); uu("raid28",d); uu("raid29",d); uu("raid20",d)
		uu("raid31",d); uu("raid32",d); uu("raid33",d); uu("raid34",d); uu("raid35",d)
		uu("raid36",d); uu("raid37",d); uu("raid38",d); uu("raid39",d); uu("raid40",d)
	elseif (event == "UNIT_PET") then
		update_unit(tagPet[arg1] or false, delta)
		if (arg1 == "player") then
			update_unit("vehicle", delta)
		end
	end
	-- if anything changed, call back
	if (deltaTail > 0) then
		return dispatch_callbacks(event)
	end
end -- process_event()

local function handle_event(self, event, arg1)
	-- if there's a callback already underway, queue the event; otherwise, process it
	if (callbackEvent) then
		eventTail = eventTail + 1
		eventQueue[eventTail] = event
		eventArg1[eventTail] = arg1
	else
		return process_event(event, arg1)
	end
end -- handle_event()

frame:SetScript("OnEvent", handle_event)


--[[ RUNTIME UPGRADE ]]--

-- undocumented, for internal use only!
function LibUnitID:__dump_for_upgrade()
	return {
		frame = frame,
		tagFunctions = tagFunctions,
		tagStaticMethods = tagStaticMethods,
		tagDynamicMethods = tagDynamicMethods,
		logActive = logActive,
		logTests = logTests,
		logUpdates = logUpdates,
		logQueries = logQueries
	}
end -- __dump_for_upgrade()

if (olddata.logActive) then
	LibUnitID:__start_logging()
end

--[[
if (olddata.tagListeners) then -- pre v.9
	olddata.tagFunctions = {}
	olddata.tagStaticMethods = {}
	olddata.tagDynamicMethods = {}
	for tag,cbs in pairs(olddata.tagListeners) do
		for cb,handler in pairs(cbs) do
			if (type(handler) == "table") then
				if (type(cb) == "function") then
					olddata.tagStaticMethods[tag] = olddata.tagStaticMethods[tag] or {}
					olddata.tagStaticMethods[tag][handler] = olddata.tagStaticMethods[tag][handler] or {}
					olddata.tagStaticMethods[tag][handler][cb] = true
				else
					olddata.tagDynamicMethods[tag] = olddata.tagDynamicMethods[tag] or {}
					olddata.tagDynamicMethods[tag][handler] = olddata.tagDynamicMethods[tag][handler] or {}
					olddata.tagDynamicMethods[tag][handler][cb] = true
				end
			else
				olddata.tagFunctions[tag] = olddata.tagFunctions[tag] or {}
				olddata.tagFunctions[tag][cb] = true
			end
		end
	end
elseif (olddata.tagMethods) then -- pre v.14
	olddata.tagStaticMethods = {}
	olddata.tagDynamicMethods = {}
	for tag,handlers in pairs(olddata.tagMethods) do
		for handler,cbs in pairs(handlers) do
			for cb,_ in pairs(cbs) do
				if (type(cb) == "function") then
					olddata.tagStaticMethods[tag] = olddata.tagStaticMethods[tag] or {}
					olddata.tagStaticMethods[tag][handler] = olddata.tagStaticMethods[tag][handler] or {}
					olddata.tagStaticMethods[tag][handler][cb] = true
				else
					olddata.tagDynamicMethods[tag] = olddata.tagDynamicMethods[tag] or {}
					olddata.tagDynamicMethods[tag][handler] = olddata.tagDynamicMethods[tag][handler] or {}
					olddata.tagDynamicMethods[tag][handler][cb] = true
				end
			end
		end
	end
end
--]]

if (olddata.tagFunctions or olddata.tagStaticMethods or olddata.tagDynamicMethods) then
	local ok
	local n,e = 0,0
	if (olddata.tagFunctions) then
		for tag,cbs in pairs(olddata.tagFunctions) do
			for cb,_ in pairs(cbs) do
				ok = add_tag_listener(tag, cb, handler)
				if (ok) then
					n = n + 1
				else
					e = e + 1
					print(MAJOR.." ERROR: Failed to re-register function callback for unit '"..tag.."' during library upgrade!")
				end
			end
			wipe(cbs)
			hashPool[cbs] = true
		end
		wipe(olddata.tagFunctions)
		hashPool[olddata.tagFunctions] = true
		olddata.tagFunctions = nil
	end
	if (olddata.tagStaticMethods) then
		for tag,handlers in pairs(olddata.tagStaticMethods) do
			for handler,cbs in pairs(handlers) do
				for cb,_ in pairs(cbs) do
					ok = add_tag_listener(tag, cb, handler)
					if (ok) then
						n = n + 1
					else
						e = e + 1
						print(MAJOR.." ERROR: Failed to re-register static method callback for unit '"..tag.."' during library upgrade!")
					end
				end
				wipe(cbs)
				hashPool[cbs] = true
			end
			wipe(handlers)
			hashPool[handlers] = true
		end
		wipe(olddata.tagStaticMethods)
		hashPool[olddata.tagStaticMethods] = true
		olddata.tagStaticMethods = nil
	end
	if (olddata.tagDynamicMethods) then
		for tag,handlers in pairs(olddata.tagDynamicMethods) do
			for handler,cbs in pairs(handlers) do
				for cb,_ in pairs(cbs) do
					ok = add_tag_listener(tag, cb, handler)
					if (ok) then
						n = n + 1
					else
						e = e + 1
						print(MAJOR.." ERROR: Failed to re-register dynamic method callback for unit '"..tag.."' during library upgrade!")
					end
				end
				wipe(cbs)
				hashPool[cbs] = true
			end
			wipe(handlers)
			hashPool[handlers] = true
		end
		wipe(olddata.tagDynamicMethods)
		hashPool[olddata.tagDynamicMethods] = true
		olddata.tagDynamicMethods = nil
	end
--[===[@debug@
	print(MAJOR..": Upgraded from revision "..olddata.MINOR.."; "..n.." callbacks re-registered, "..e.." errors.")
elseif (olddata.MINOR) then
	print(MAJOR..": Upgraded from revision "..olddata.MINOR.."; no callbacks re-registered.")
--@end-debug@]===]
end

wipe(olddata)
hashPool[olddata] = true
olddata = nil


--[[ PUBLIC API METHODS ]]--

--=== compatibility ===--

--- Gets a list of the UnitIDs that the library is able to track and cache.
-- The returned table contains the supported UnitIDs as both keys and values,
-- such as '''{target="target",pet="pet",...}'''.  These are the UnitIDs that
-- may be registered for callbacks using ''':RegisterUnit()''', and then
-- queried using ''':GetUnitGUID()''' or '''.UnitGUID()'''.
-- @return A hash table containing the supported UnitIDs.
function LibUnitID:GetSupportedUnitIDs()
	-- build a fresh table, in case the caller wants to muck it up
	local t = next(hashPool) or {}
	hashPool[t] = nil
	for tag,_ in pairs(tagSupported) do
		t[tag] = tag
	end
	return t
end -- GetSupportedUnitIDs()

--- Tests if the specified UnitID is supported by the library.
-- @param unitid The UnitID to query support for.
-- @return True if the UnitID can be tracked and cached by the library, false otherwise.
function LibUnitID:IsUnitIDSupported(unitid)
	return (tagSupported[unitid] and true) or false
end -- IsUnitIDSupported()

--=== register unit ===--

--- Registers a callback for a single UnitID.
-- Whenever the identity of the specified UnitID changes, the callback
-- function will be called.  If a handler is provided, it must be a table and
-- will be given as the first argument to the callback.  The next arguments to
-- the callback will be the name of the game event that triggered the update,
-- followed by a '''delta''' table detailing the UnitIDs that changed.
--
-- If a handler is provided, then '''callback''' may be specified as a string,
-- which will be resolved on the handler each time the callback is to be
-- called.  This is slightly slower but more flexible, since it allows you to
-- redefine the callback function without having to unregister and re-register
-- the function pointer.
--
-- Multiple calls with the same '''unitid''', '''callback''' and '''handler'''
-- have no effect.  A single '''callback''' may be registered with multiple
-- '''handler'''s, and a single '''handler''' may be registered with multiple
-- '''callback'''s, and each combination will be called.
--
-- If the same '''callback'''+'''handler''' pair is registered for multiple
-- UnitIDs, and if a single game event causes more than one of those UnitIDs
-- to change identity, the callback will still only be called once.  To
-- retrieve information about which UnitIDs actually changed within your
-- callback, check the provided '''delta''' table.
-- @see delta
-- @param unitid The UnitID to monitor.
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return True if the callback is successfully registered, false otherwise.
function LibUnitID:RegisterUnit(unitid, callback, handler)
	return add_tag_listener(unitid, callback, handler)
end -- RegisterUnit()

--=== register party ===--

--- Registers a callback for all party-member UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs "party1"
-- through "party5".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterParty(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("party"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterParty()

--- Registers a callback for all party-member-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "party1target" through "party5target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterPartyTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("party"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyTargets()

--- Registers a callback for all party-member-pet UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "partypet1" through "partypet5".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterPartyPets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("partypet"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyPets()

--- Registers a callback for all party-member-pet-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "partypet1target" through "partypet5target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (5 possible).
function LibUnitID:RegisterPartyPetTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((add_tag_listener("partypet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterPartyPetTarges()

--- Registers a callback for all party related UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "party#", "party#target", "partypet#" and "partypet#target" for # 1..5.
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (20 possible).
function LibUnitID:RegisterAllParty(callback, handler)
	local ok = 0
	for tag,_ in pairs(partyTags) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAllParty()

--=== register raid ===--

--- Registers a callback for all raid-member UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs "raid1"
-- through "raid40".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaid(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raid"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaid()

--- Registers a callback for all raid-member-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raid1target" through "raid40target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaidTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raid"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidTargets()

--- Registers a callback for all raid-member-pet UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raidpet1" through "raidpet40".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaidPets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raidpet"..i, callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidPets()

--- Registers a callback for all raid-member-pet-target UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raidpet1target" through "raidpet40target".
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (40 possible).
function LibUnitID:RegisterRaidPetTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((add_tag_listener("raidpet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- RegisterRaidPetTargets()

--- Registers a callback for all raid related UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on the UnitIDs
-- "raid##", "raid##target", "raidpet##" and "raidpet##target" for ## 1..40.
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered (160 possible).
function LibUnitID:RegisterAllRaid(callback, handler)
	local ok = 0
	for tag,_ in pairs(raidTags) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAllRaid()

--=== register all ===--

--- Registers a callback for all supported UnitIDs.
-- This is equivalent to calling ''':RegisterUnit()''' on all UnitIDs that
-- are supported by LibUnitID.
-- @see LibUnitID:RegisterUnit
-- @param callback A function reference to call, or a string method name to resolve on the '''handler''' and then call.
-- @param handler (Optional) The first argument to pass to the callback function, and the table to resolve '''callback''' on if it is a string.
-- @return The number of UnitIDs that were successfully registered.
function LibUnitID:RegisterAll(callback, handler)
	local ok = 0
	for tag,_ in pairs(tagSupported) do ok = ok + ((add_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- RegisterAll()

--=== unregister unit ===--

--- Unregisters a callback from a single UnitID.
-- The callback function or method will no longer be called for changes in the
-- specified UnitID.  Callbacks are stored separately and must be unregistered
-- separately for each '''callback'''+'''handler''' pair, including whether
-- '''callback''' is a function reference or a string method name.
-- @param unitid The UnitID to stop monitoring.
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return True if the callback is successfully unregistered, false otherwise (including if the callback was not registered for the unitid).
function LibUnitID:UnregisterUnit(unitid, callback, handler)
	return remove_tag_listener(unitid, callback, handler)
end -- UnregisterUnit()

--=== unregister party ===--

--- Unregisters a callback from all party-member UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "party1" through "party5".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterParty(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("party"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterParty()

--- Unregisters a callback from all party-member-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "party1target" through "party5target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterPartyTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("party"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyTargets()

--- Unregisters a callback from all party-member-pet UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "partypet1" through "partypet5".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterPartyPets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("partypet"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyPets()

--- Unregisters a callback from all party-member-pet-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "partypet1target" through "partypet5target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (5 possible).
function LibUnitID:UnregisterPartyPetTargets(callback, handler)
	local ok = 0
	for i=1,5 do ok = ok + ((remove_tag_listener("partypet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterPartyPetTargets()

--- Unregisters a callback from all party related UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "party#", "party#target", "partypet#" and "partypet#target" for # 1..5.
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (20 possible).
function LibUnitID:UnregisterAllParty(callback, handler)
	local ok = 0
	for tag,_ in pairs(partyTags) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAllParty()

--=== unregister raid ===--

--- Unregisters a callback from all raid-member UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raid1" through "raid40".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaid(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raid"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaid()

--- Unregisters a callback from all raid-member-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raid1target" through "raid40target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaidTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raid"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidTargets()

--- Unregisters a callback from all raid-member-pet UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raidpet1" through "raidpet40".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaidPets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raidpet"..i, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidPets()

--- Unregisters a callback from all raid-member-pet-target UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raidpet1target" through "raidpet40target".
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (40 possible).
function LibUnitID:UnregisterRaidPetTargets(callback, handler)
	local ok = 0
	for i=1,40 do ok = ok + ((remove_tag_listener("raidpet"..i.."target", callback, handler) and 1) or 0) end
	return ok
end -- UnregisterRaidPetTargets()

--- Unregisters a callback from all raid related UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on the UnitIDs
-- "raid#", "raid#target", "raidpet#" and "raidpet#target" for # 1..40.
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered (160 possible).
function LibUnitID:UnregisterAllRaid(callback, handler)
	local ok = 0
	for tag,_ in pairs(raidTags) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAllRaid()

--=== unregister all ===--

--- Unregisters a callback from all supported UnitIDs.
-- This is equivalent to calling ''':UnregisterUnit()''' on all UnitIDs that
-- are supported by LibUnitID.
-- @see LibUnitID:UnregisterUnit
-- @param callback A function reference or string method name to stop calling.
-- @param handler (Optional) The handler associated with the callback to unregister.
-- @return The number of UnitIDs that were successfully unregistered.
function LibUnitID:UnregisterAll(callback, handler)
	local ok = 0
	for tag,_ in pairs(tagSupported) do ok = ok + ((remove_tag_listener(tag, callback, handler) and 1) or 0) end
	return ok
end -- UnregisterAll()

--=== guid queries ===--

--- Returns the UnitID's GUID from the cache.
-- This is the method-call version of this functionality; it must be called as
-- a method on a LibUnitID library reference, as in
-- '''myLibRef:GetUnitGUID("pet")'''.
--
-- Note that the cache is only updated when the library receives event
-- notifications from the game engine, which may happen in any order.  If you
-- have your own event handlers for events that affect UnitIDs, and if you try
-- to query the cache during those event handlers, you may receive out-of-date
-- information if the library has not yet received the event and updated
-- itself.
--
-- Also note that the library will not track UnitIDs for which nobody has
-- registered a callback, and will also not take the time to verify that when
-- queried; if you request the GUID of a UnitID that you did not register a
-- callback for, you will get '''nil''' unless someone else happened to
-- register for that UnitID.
--
-- For best results, only query the cache from a callback that you registered
-- with LibUnitID, or during events that do not affect UnitIDs.
-- @param unitid The UnitID to fetch the GUID for.
-- @return The GUID of the UnitID, or nil if it doesn't exist or isn't being tracked by the library.
function LibUnitID:GetUnitGUID(unitid)
	return tagID[unitid]
end -- GetUnitGUID()

--- Returns the UnitID's GUID from the cache.
-- This is the function-call version of this functionality; it must be called
-- as a bare function, as in '''myLibRef.UnitGUID("pet")'''.  This is done so
-- that you can make a local copy of the function reference (as in
-- '''local UnitGUID = myLibRef.UnitGUID''') and then use it in place of the
-- default API UnitGUID() function.  In all other ways this is equivalent to
-- ''':GetUnitGUID()''', and all the same caveats apply.
-- @see LibUnitID:GetUnitGUID
-- @param unitid The UnitID to fetch the GUID for.
-- @return The GUID of the UnitID, or nil if it doesn't exist or isn't being tracked by the library.
function LibUnitID.UnitGUID(unitid) -- NOT A METHOD!  that's a dot, not a colon, Jim!
	return tagID[unitid]
end -- UnitGUID()


--[[ STATISTICS & ANALYSIS ]]--
-- undocumented, but needed by the disembedded utilities

local real_update_unit = update_unit
local function log_update_unit(tag, delta)
	logTests = logTests + 1
	return real_update_unit(tag, delta)
end -- log_update_unit()

local real_dispatch_callbacks = dispatch_callbacks
local function log_dispatch_callbacks(event)
	logUpdates = logUpdates + deltaTail
	return real_dispatch_callbacks(event)
end -- log_dispatch_callbacks()

local real_UnitGUID = LibUnitID.UnitGUID
local function log_UnitGUID(unitid)
	logQueries = logQueries + 1
	return real_UnitGUID(unitid)
end -- log_UnitGUID()

local real_GetUnitGUID = LibUnitID.GetUnitGUID
local function log_GetUnitGUID(self, unitid)
	logQueries = logQueries + 1
	return real_GetUnitGUID(self, unitid)
end -- log_GetUnitGUID()

function LibUnitID:__start_logging()
	logActive = true
	update_unit = log_update_unit
	dispatch_callbacks = log_dispatch_callbacks
	self.GetUnitGUID = log_GetUnitGUID
	self.UnitGUID = log_UnitGUID
end -- __start_logging()

function LibUnitID:__get_statistics()
	return logTests,logUpdates,logQueries
end -- __get_statistics()

function LibUnitID:__reset_statistics()
	logTests = 0
	logUpdates = 0
	logQueries = 0
end -- __reset_statistics()

function LibUnitID:__stop_logging()
	logActive = false
	update_unit = real_update_unit
	dispatch_callbacks = real_dispatch_callbacks
	self.GetUnitGUID = real_GetUnitGUID
	self.UnitGUID = real_UnitGUID
end -- __stop_logging()

function LibUnitID:__get_speed_tests()
	return compile_update_unit_tests()
end -- __get_speed_tests()

