local MAJOR = "LibUnitID-1.0"
--@alpha@
MAJOR = MAJOR.."-alpha"
--@end-alpha@
local MINOR = 14
local REQMINOR = 14
local LibUnitID,libminor = LibStub(MAJOR)

-- set a default chat command hook in case we bail out
SlashCmdList["LIBUNITID"] = function()
	print("LibUnitID: Utilities rev. "..MINOR.." are not compatible with library rev. "..libminor..".")
end
SLASH_LIBUNITID1 = "/libunitid"
SLASH_LIBUNITID2 = "/lui"

-- bail out if the library core is older than supported or newer than these utilities
if (libminor < REQMINOR or libminor > MINOR) then
	SlashCmdList["LIBUNITID"]()
	return
end

-- check which addons already loaded, and thus probably already localized UnitGUID
local preloaded = {}
do
	local n = GetNumAddOns()
	local name,loaded
	if (n > 0) then
		for i=1,n do
			name = (GetAddOnInfo(i)) -- name,title,notes,enabled,loadable,reason,security
			loaded = (IsAddOnLoaded(i)) -- toc,ADDON_LOADED
			if (name ~= "!LibUnitID" and loaded) then
				preloaded[#preloaded+1] = name
				preloaded[name] = true
			end
		end
	end
end

-- initialize vars
local postloaded
local monitorCount = 0
local timingRuns = 0
local timeAPI = 0
local timeCacheTest = 0
local timeCacheUpdate = 0
local timeCacheHit = 0
local logSince = GetTime()
local logCachableAPI = 0
local logUncachableAPI = { [0]=0 }

-- get the supported UnitIDs and speed test functions from the library
local tagSupported = LibUnitID:GetSupportedUnitIDs()
local libCacheTest,libCacheUpdate = LibUnitID:__get_speed_tests()

-- store initial function pointers
local API_UnitGUID = UnitGUID
local lib_UnitGUID = LibUnitID.UnitGUID

-- set the trap (the relay mechanism is for rascally addons who localize function pointers)
local relay_API_UnitGUID = API_UnitGUID
UnitGUID = function(...) return relay_API_UnitGUID(...) end
_G["UnitGUID"] = function(...) return relay_API_UnitGUID(...) end

-- define the wrapper
local log_API_UnitGUID = function(unitid, ...)
	if (tagSupported[unitid]) then
		logCachableAPI = logCachableAPI + 1
	else
		logUncachableAPI[0] = logUncachableAPI[0] + 1
		logUncachableAPI[unitid] = (logUncachableAPI[unitid] or 0) + 1
	end
	return API_UnitGUID(unitid, ...)
end

-- monitor/eventhandler frame
local frame = CreateFrame("Frame", nil, UIParent)

local monitor_handler = {}

monitor_handler.monitor_callback = function(self, event, delta)
	if (self ~= monitor_handler) then
		print("LibUnitID: Bad method callback for " .. event)
	end
	if (LibUnitID_DB.verbose) then
		local n = delta[0]
		local s = ""
		for i=1,n do
			s = s..", "..delta[i].."("..tostring(lib_UnitGUID(delta[i]))..")"
		end
		if (s == "") then
			print("LibUnitID: "..event.." => EMPTY UPDATE")
		else
			print("LibUnitID: "..event.." => "..strsub(s,3))
		end
	end
end -- monitor_callback()

local function monitor_OnShow()
	-- register both ways as a test; we should still only see one call per event
	monitorCount = LibUnitID:RegisterAll(monitor_handler.monitor_callback, monitor_handler)
	if (monitorCount < 1) then
		print("LibUnitID: Failed to register monitor handler function.")
	end
	local altCount = LibUnitID:RegisterAll("monitor_callback", monitor_handler)
	if (altCount ~= monitorCount) then
		print("LibUnitID: Failed to register monitor handler method.")
	end
end -- monitor_OnShow()

local function monitor_OnHide()
	local unCount = LibUnitID:UnregisterAll(monitor_handler.monitor_callback, monitor_handler)
	if (unCount ~= monitorCount) then
		print("LibUnitID: Failed to unregister monitor handler function.")
	end
	unCount = LibUnitID:UnregisterAll("monitor_callback", monitor_handler)
	if (unCount ~= monitorCount) then
		print("LibUnitID: Failed to unregister monitor handler method.")
	end
	monitorCount = 0
end -- monitor_OnHide()

-- when savedvars load, initialize logging and monitoring
frame:SetScript("OnEvent", function(self, event, arg1)
	if (arg1 == "!LibUnitID") then
		if (not LibUnitID_DB) then
			LibUnitID_DB = {}
		end
		if (LibUnitID_DB.log) then
			relay_API_UnitGUID = log_API_UnitGUID
			LibUnitID:__start_logging()
		end
		frame:UnregisterEvent("ADDON_LOADED")
		frame:SetScript("OnEvent", nil)
		frame:Hide()
		frame:SetScript("OnShow", monitor_OnShow)
		frame:SetScript("OnHide", monitor_OnHide)
		if (LibUnitID_DB.monitor) then
			frame:Show()
--[===[@debug@
			-- test availability of UnitGUID(player)
			if (API_UnitGUID("player")) then
				print("LibUnitID: UnitGUID(player) available at ADDON_LOADED(!LibUnitID)")
			else
				print("LibUnitID: UnitGUID(player) not yet available at ADDON_LOADED(!LibUnitID)")
				local updates = 0
				frame:SetScript("OnUpdate", function()
					updates = updates + 1
					if (API_UnitGUID("player")) then
						print("LibUnitID: UnitGUID(player) available after "..updates.." OnUpdate frames")
						frame:SetScript("OnUpdate", nil)
					end
				end)
			end
--@end-debug@]===]
		end
	end
end)
frame:RegisterEvent("ADDON_LOADED")


--[[ TIMING TEST ]]--

local function generate_timing_data()
	local GetTime = GetTime
	local newRuns = timingRuns + 1
	local n,t,func
	-- sanity check
	if (not API_UnitGUID("target")) then return false end
	-- timing loop overhead
	n = 128
	repeat
		n = n * 2
		t = GetTime()
		for i=1,n do end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	local timeLoop = (t * 1e9) / n -- nanoseconds per loop; not refined: we always want a current snapshot
	-- API call
	n = 128
	repeat
		local func = API_UnitGUID
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeAPI = ((timeAPI * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- library cache test
	n = 128
	repeat
		local func = libCacheTest
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheTest = ((timeCacheTest * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- library cache update
	n = 128
	repeat
		local func = libCacheUpdate
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheUpdate = ((timeCacheUpdate * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- library cache hit
	n = 128
	repeat
		local func = lib_UnitGUID
		n = n * 2
		t = GetTime()
		for i=1,n do func("target") end
		t = GetTime() - t
	until (n >= 2e9 or t > 0.2)
	timeCacheHit = ((timeCacheHit * timingRuns) + (  ((t * 1e9) / n) - timeLoop  )) / newRuns -- nanoseconds per loop
	-- done
	timingRuns = newRuns
	return true
end -- generate_timing_data()


--[[ COMMAND PARSER ]]--

local function s(f) return (f<0.001 and format("%.3f",f*1e6).."us") or (f<1 and format("%.3f",f*1e3).."ms") or format("%.3f",f).."s" end
local function ms(f) return (f<1 and format("%.3f",f*1e3).."us") or (f<1e3 and format("%.3f",f).."ms") or format("%.3f",f/1e3).."s" end
local function us(f) return (f<1e3 and format("%.3f",f).."us") or (f<1e6 and format("%.3f",f/1e3).."ms") or format("%.3f",f/1e6).."s" end
local function ns(f) return (f<1e3 and format("%.3f",f).."ns") or (f<1e6 and format("%.3f",f/1e3).."us") or format("%.3f",f/1e6).."ms" end
local function pct(f) return format("%.3f",f).."%" end

local function do_list(args, notag)
	if (#preloaded == 0) then
		print((notag and "") or "LibUnitID: ", "No addons loaded before !LibUnitID; all calls should be trapped.")
	else
		print((notag and "") or "LibUnitID: ", #preloaded.." addons loaded before !LibUnitID, and their GUID queries (both native-API and library-cached) might bypass the logging hooks:")
		local s = ""
		for n,name in ipairs(preloaded) do
			s = s..", "..name
		end
		print("  "..strsub(s,3))
	end
	if (args == "all") then
		if (not postloaded) then
			postloaded = {}
			local n = GetNumAddOns()
			local name,loaded
			for i=1,n do
				name = (GetAddOnInfo(i)) -- name,title,notes,enabled,loadable,reason,security
				loaded = (IsAddOnLoaded(i)) -- toc,ADDON_LOADED
				if (name ~= "!LibUnitID" and loaded and not preloaded[name]) then
					postloaded[#postloaded+1] = name
					postloaded[name] = true
				end
			end
		end
		print(#postloaded.." addons loaded after !LibUnitID:")
		local s = ""
		for n,name in ipairs(postloaded) do
			s = s..", "..name
		end
		print("  "..strsub(s,3))
	end
end -- do_list()

local function do_log(args, notag)
	if (args == "on") then
		if (LibUnitID_DB.log) then
			print((notag and "") or "LibUnitID: ", "Logging is already on.")
		else
			relay_API_UnitGUID = log_API_UnitGUID
			LibUnitID:__start_logging()
			LibUnitID_DB.log = true
			print((notag and "") or "LibUnitID: ", "GUID queries and library cache upkeep are now logged.")
			if (#preloaded > 0) then print("LibUnitID: "..#preloaded.." addons loaded before !LibUnitID, and their GUID queries might bypass the logging hooks; check '/libunitid list'.") end
		end
	elseif (args == "off") then
		if (LibUnitID_DB.log) then
			relay_API_UnitGUID = API_UnitGUID
			LibUnitID:__stop_logging()
			LibUnitID_DB.log = nil
			print((notag and "") or "LibUnitID: ", "GUID queries and library cache upkeep are no longer logged.")
		else
			print((notag and "") or "LibUnitID: ", "Logging is already off.")
		end
	else
		print((notag and "") or "LibUnitID: ", "Logging is currently "..((LibUnitID_DB.log and "ON") or "OFF")..".")
	end
end -- do_log()

local function do_monitor(args, notag)
	if (args == "on") then
		if (LibUnitID_DB.monitor) then
			print((notag and "") or "LibUnitID: ", "Monitoring is already on.")
		else
			LibUnitID_DB.monitor = true
			frame:Show()
			print((notag and "") or "LibUnitID: ", "All supported UnitIDs are now monitored ("..monitorCount.." UnitIDs).")
		end
	elseif (args == "off") then
		if (LibUnitID_DB.monitor) then
			LibUnitID_DB.monitor = nil
			frame:Hide()
			print((notag and "") or "LibUnitID: ", "Monitoring is now disabled.")
		else
			print((notag and "") or "LibUnitID: ", "Monitoring is already off.")
		end
	else
		print((notag and "") or "LibUnitID: ", "Monitoring is currently "..((LibUnitID_DB.monitor and "ON") or "OFF").." ("..monitorCount.." UnitIDs monitored).")
	end
end -- do_monitor()

local function do_verbose(args, notag)
	if (args == "on") then
		if (LibUnitID_DB.verbose) then
			print((notag and "") or "LibUnitID: ", "Monitoring is already verbose.")
		else
			LibUnitID_DB.verbose = true
			print((notag and "") or "LibUnitID: ", "Monitoring is now verbose.")
		end
	elseif (args == "off") then
		if (LibUnitID_DB.verbose) then
			LibUnitID_DB.verbose = nil
			print((notag and "") or "LibUnitID: ", "Monitoring is no longer verbose.")
		else
			print((notag and "") or "LibUnitID: ", "Monitoring is already not verbose.")
		end
	else
		print((notag and "") or "LibUnitID: ", "Verbose monitoring mode is currently "..((LibUnitID_DB.verbose and "ON") or "OFF")..".")
	end
end -- do_verbose()

local function do_time(calc, notag)
	if (calc) then
		print((notag and "") or "LibUnitID: ", "Generating timing data...")
		if (generate_timing_data()) then
			print("...done.  Results aggregated over "..timingRuns.." run"..((timingRuns>1 and "s") or "")..".")
		else
			print("...aborted: You must have a target to generate timing data; target something and try again.")
			return
		end
	end
	print("Nanoseconds per operation:")
	print("  "..ns(timeAPI).." per API UnitGUID()")
	print("  "..ns(timeCacheTest).." per Cache Change Test")
	print("  "..ns(timeCacheUpdate).." per Cache Identity Update")
	print("  "..ns(timeCacheHit).." per Cached GUID Query")
end -- do_time()

local function do_report(args, notag)
	local interval = GetTime() - logSince -- seconds
	local logCacheTests,logCacheUpdates,logCacheQueries = LibUnitID:__get_statistics()
	print((notag and "") or "LibUnitID: ", "Operations in the last "..s(interval)..":")
	print("  "..logCachableAPI.." Cachable API UnitGUID()s")
	print("  "..logUncachableAPI[0].." Uncachable API UnitGUID()s")
	print("  "..logCacheTests.." Cache Change Tests")
	print("  "..logCacheUpdates.." Cache Identity Updates")
	print("  "..logCacheQueries.." Cached GUID Queries")
	if (args == "all") then
		print("Uncachable UnitGUID() queries:")
		for tag,count in pairs(logUncachableAPI) do
			if (tag ~= 0) then
				print("  "..tag.." - "..count)
			end
		end
	end
end -- do_report()

local function do_analyze(args, notag)
	local interval = GetTime() - logSince -- seconds
	local logCacheTests,logCacheUpdates,logCacheQueries = LibUnitID:__get_statistics()
	if (logCachableAPI < 1 and logUncachableAPI < 1 and logCacheTests < 1 and logCacheUpdates < 1) then
		print((notag and "") or "LibUnitID: ", "Efficiency analysis not available without logging data; run '/libunitid log on'")
		return
	elseif (timingRuns < 1) then
		print((notag and "") or "LibUnitID: ", "Efficiency analysis not available without timing data; run '/libunitid time'")
		return
	end
	do_time(false)
	print((notag and "") or "LibUnitID: ", "Analysis:")
	local api = ((logCachableAPI * timeAPI) / 1e3) + ((logUncachableAPI[0] * timeAPI) / 1e3) -- milliseconds
	print("  "..ms(api).." spent querying the API's UnitGUID() ("..logCachableAPI.." cachable, "..logUncachableAPI[0].." uncachable)")
	local cost = ((logCacheTests * timeCacheTest) / 1e3) + ((logCacheUpdates * timeCacheUpdate) / 1e3) -- milliseconds
	print("  "..ms(cost).." spent maintaining the library's cache ("..logCacheTests.." tests, "..logCacheUpdates.." updates)")
	local benefit = ((logCacheQueries * timeAPI) / 1e3) - ((logCacheQueries * timeCacheHit) / 1e3) -- milliseconds
	print("  "..ms(benefit).." saved by querying the cache instead of the API ("..logCacheQueries.." queries)")
	local net = benefit - cost -- milliseconds
	local efficiency = ((net / 1e3) * 100.0) / interval
	if (net < 0) then
		print("Overall loss is "..pct(efficiency).." ("..ms(-net).." extra spent over the last "..s(interval)..")")
	else
		print("Overall gain is "..pct(efficiency).." ("..ms(net).." saved over the last "..s(interval)..")")
	end
	local bestBenefit = ((logCachableAPI * timeAPI) / 1e3) - ((logCachableAPI * timeCacheHit) / 1e3) -- milliseconds
	print("  "..ms(bestBenefit).." more could have been saved if cachable UnitGUID() calls had used the cache")
	local bestNet = net + bestBenefit -- milliseconds
	local bestEfficiency = ((bestNet / 1e3) * 100.0) / interval
	if (bestNet < 0) then
		print("Best-case loss is "..pct(bestEfficiency).." ("..ms(-bestNet).." extra still spent over the last "..s(interval)..")")
	else
		print("Best-case gain is "..pct(bestEfficiency).." ("..ms(bestNet).." could have been saved over the last "..s(interval)..")")
	end
end -- do_analyze()

SlashCmdList["LIBUNITID"] = function(args)
	-- bail out if the library core is older than supported or newer than these utilities
	local LibUnitID,libminor = LibStub(MAJOR)
	if (libminor < REQMINOR or libminor > MINOR) then
		print("LibUnitID: Utilities rev. "..MINOR.." are not compatible with library rev. "..libminor..".")
		return
	end
	-- process command
	if (strsub(args,1,4) == "list") then
		do_list(strsub(args,6))
	elseif (strsub(args,1,3) == "log") then
		do_log(strsub(args,5))
	elseif (strsub(args,1,7) == "monitor") then
		do_monitor(strsub(args,9))
	elseif (strsub(args,1,7) == "verbose") then
		do_verbose(strsub(args,9))
	elseif (strsub(args,1,4) == "time") then
		do_time(true)
    elseif (strsub(args,1,6) == "report") then
		do_report(strsub(args,8))
    elseif (strsub(args,1,7) == "analyze") then
		do_analyze(strsub(args,9))
	elseif (strsub(args,1,5) == "reset") then
		timingRuns = 0
		timeAPI = 0
		timeCacheTest = 0
		timeCacheUpdate = 0
		timeCacheHit = 0
		logSince = GetTime()
		logCachableAPI = 0
		logUncachableAPI = { [0]=0 }
		LibUnitID:__reset_statistics()
		print("LibUnitID: Statistics reset.")
	elseif (args == "help") then
		print("LibUnitID commands via '/libunitid' or '/lui':")
		print("  /libunitid list [all]  --  list addons that loaded before (and optionally after) !LibUnitID")
		print("  /libunitid log [on | off]  --  log calls to API's and library's UnitGUID()")
		print("  /libunitid monitor [on | off]  --  monitor all supported UnitIDs (forces maximum cache upkeep work)")
		print("  /libunitid verbose [on | off]  --  when monitoring, also print all identity changes (spammy!)")
		print("  /libunitid time  --  generate or refine timing data for performance analysis")
		print("  /libunitid report [all]  --  show logging results")
		print("  /libunitid analyze  --  show performance analysis (requires log and timing data)")
		print("  /libunitid reset  --  reset call log counters and timing data")
		print("  /libunitid help  --  this")
    else
		print(MAJOR.." standalone utilities rev. "..MINOR)
		do_log(nil, true)
		do_monitor(nil, true)
		do_verbose(nil, true)
		print("'/libunitid help' for usage.")
    end
end

